"""
    picasso/__init__.py
    ~~~~~~~~~~~~~~~~~~~~

    :authors: Joerg Schnitzbauer, Maximilian Thomas Strauss, Rafal Kowalewski 2016-2023
    :copyright: Copyright (c) 2016-2023 Jungmann Lab, MPI of Biochemistry
"""
import os.path as _ospath
import yaml as _yaml

__version__ = "0.6.8"

_this_file = _ospath.abspath(__file__)
_this_dir = _ospath.dirname(_this_file)
try:
    with open(_ospath.join(_this_dir, "config.yaml"), "r") as config_file:
        CONFIG = _yaml.full_load(config_file)
    if CONFIG is None:
        CONFIG = {}
except FileNotFoundError:
    CONFIG = {}


import streamlit as st
import os
import pandas as pd
from multiprocessing import Process
import time
from sqlalchemy import create_engine
from picasso import localize
from helper import fetch_watcher, fetch_db
import psutil
import subprocess
from datetime import datetime
import sys

DEFAULT_UPDATE_TIME = 1

FILETYPES = (".raw", ".ome.tif", ".ims")
from picasso.__main__ import _localize


class aclass:
    def __init__(self, **entries):
        self.__dict__.update(entries)


def check_new(path: str, processed: dict, logfile: str):
    """Check if files in a folder are not processed yet.
    Files are considered processed if they have a _locs.hdf5 file.

    Args:
        path (str): Folder to check.
        processed (dict): Dict of files that are already processed.
        logfile (str): Path to logfile.

    Returns:
        _type_: _description_
    """

    all_ = os.listdir(path)
    all_ = [os.path.join(path, _) for _ in all_]

    new = [_ for _ in all_ if os.path.normpath(_) not in processed.keys() and _.endswith(FILETYPES)]
    locs = [_ for _ in all_ if _.endswith("_locs.hdf5")]

    print_to_file(
        logfile,
        f"{datetime.now()} Checking: {len(all_)} files, {len(new)} unprocessed with valid ending and {len(locs)} `_locs.hdf5` files in {path}.",
    )

    for _ in new:
        base, ext = os.path.splitext(_)
        for ref in locs:
            if base.startswith(ref):
                processed[_] = True
                new.remove(_)
                break

    return new, processed


def wait_for_change(file: str):
    """Helper function that checks if a file is changing the size.

    Args:
        file (str): Path to file.
    """
    print(f"Waiting for {file}")
    filesize = os.path.getsize(file)
    writing = True
    while writing:
        time.sleep(2)
        new_filesize = os.path.getsize(file)
        if filesize == new_filesize:
            writing = False
        else:
            filesize = new_filesize

    print(f"File {file} complete.")


def get_children_files(file: str, checked: list):
    """Helper function that extracts files with the same start and same ending.

    Args:
        file (str): Path to check.
        checked (list): List with files that are already checked.
    """
    dir_ = os.path.dirname(file)
    files_in_folder = [os.path.abspath(os.path.join(dir_, _)) for _ in os.listdir(dir_)]
    # Multiple ome tifs; Pos0.ome.tif', Pos0_1.ome.tif', Pos0_2.ome.tif'
    files_in_folder = [
        _
        for _ in files_in_folder
        if _.startswith(file[:-8]) and _ not in checked and _.endswith(".ome.tif") and 'MMStack_Pos0' in _
    ]

    for _ in files_in_folder:
        wait_for_change(_)

    return files_in_folder


def wait_for_completion(file: str):
    """Helper function that waits until a file is completely written.

    Args:
        file (str): Filepath.
    """

    wait_for_change(file)

    if file.endswith(".ome.tif"):
        checked = [file]

        time.sleep(2)

        children = get_children_files(file, checked)
        checked.extend(children)

        while len(children) > 0:
            children = get_children_files(file, checked)
            checked.extend(children)

    else:
        checked = []

    return checked


def print_to_file(path, text):
    with open(path, "a") as f:
        f.write(text)
        f.write("\n")


def check_new_and_process(
    settings_list: dict, path: str, command: str, logfile: str, existing: list, update_time: int
):
    """
    Checks a folder for new files and processes them with defined settigns.
    Args:
        settings_list (list): List of dictionaries with settings.
        path (str): Path to folder.
        command (str): Command to execute after processing.
        logfile (str): Path to logfile.
        existing (list): existing files 
        update_time (int): Refresh every x minutes
    """

    print_to_file(logfile, f"{datetime.now()} Started watcher for {path}.")
    print_to_file(logfile, f"{datetime.now()} Settings {settings_list}.")

    processed = {}

    for _ in existing:
        processed[_] = True

    while True:
        new, processed = check_new(path, processed, logfile)

        if len(new) > 0:
            file = os.path.abspath(new[0])
            print_to_file(logfile, f"{datetime.now()} New file {file}")
            children = wait_for_completion(file)
            print_to_file(logfile, f"{datetime.now()} Children {children}")
            try:
                for settings in settings_list:

                    if len(settings_list) > 1:
                        print_to_file(
                            logfile,
                            f"{datetime.now()} Processing group {settings['suffix']}",
                        )

                    settings["files"] = file

                    args_ = aclass(**settings)
                    _localize(args_)

                if command != "":

                    if "$FILENAME" in command:
                        to_execute = command[:]
                        to_execute = to_execute.replace("$FILENAME", f'"{file}"')

                    print_to_file(logfile, f"{datetime.now()} Executing {to_execute}.")

                    subprocess.run(to_execute)

            except KeyboardInterrupt:
                raise
            except Exception as e:
                print_to_file(logfile, f"{datetime.now()} Exception {e} occured.")

            processed[file] = True

            for _ in children:
                processed[_] = True

            print_to_file(logfile, f"{datetime.now()} File {file} processed.")
            print(f"{datetime.now()} File {file} processed.")

        time.sleep(update_time * 60)


def watcher():
    """
    Streamlit page to show the watcher page.
    """
    st.write("# Watcher")
    st.text(
        "- Set up a file watcher to process files in a folder with pre-defined settings automatically."
    )
    st.text(
        "- All new files and raw files that aren't yet in the database will be processed."
    )
    st.text(
        "- You can define different parameter groups so that a file will be processed with different settings."
    )
    st.text("- You can also chain custom commands to the watcher.")
    st.text(f"- The watcher will check for the following filetypes: {FILETYPES}")

    st.write("## Existing watchers")
    df_ = fetch_watcher()
    df = df_.copy()
    if len(df) > 0:
        df["running"] = [psutil.pid_exists(_) for _ in df["process id"]]
        st.dataframe(df)
        if df["running"].sum() != len(df):
            if st.button("Remove non-running watchers."):
                df = df[df["running"]]
                engine = create_engine(
                    "sqlite:///" + localize._db_filename(), echo=False
                )
                df.to_sql("watcher", con=engine, if_exists="replace", index=False)

                st.success("Removed. Please refresh this page.")

                st.stop()

        logfile = st.selectbox("Select logfile", df["logfile"].iloc[::-1])
        with st.expander("Log"):
            with open(logfile, "r") as f:
                text = f.readlines()
                st.text("".join(text))

    else:
        st.write("None")

    st.write("## New watcher")
    folder = st.text_input("Enter folder to watch.", os.getcwd())

    if not os.path.isdir(folder):
        st.error("Not a valid path.")
    else:
        with st.expander("Settings"):

            n_columns = int(
                st.number_input(
                    "Number of Parameter Groups", min_value=1, max_value=10, step=1
                )
            )

            if n_columns > 1:
                st.text(
                    "Parameter groups will be indiciated with a `pg` in the filename, e.g. `filename_pg_1_locs.hdf`"
                )

            columns = st.columns(n_columns)
            settings = {}

            for i in range(n_columns):
                col = columns[i]
                settings[i] = {}
                if n_columns > 1:
                    col.write(f"Group {i+1} (pg_{i+1})")

                settings[i]["box"] = col.number_input(
                    label="Box side length:", value=7, key=f"box_{i}"
                )
                settings[i]["min_net_gradient"] = col.number_input(
                    label="Min. Net Gradient:",
                    value=5000,
                    min_value=0,
                    max_value=10000,
                    key=f"gradient_{i}",
                )

                settings[i]["em_gain"] = col.number_input(
                    label="EM Gain:",
                    value=1,
                    min_value=1,
                    max_value=1000,
                    key=f"em_{i}",
                )

                settings[i]["baseline"] = col.number_input(
                    label="Baseline:", value=100, key=f"baseline_{i}"
                )
                settings[i]["sensitivity"] = col.number_input(
                    label="Sensitivity:", value=1.00, key=f"sensitivity_{i}"
                )
                settings[i]["qe"] = col.number_input(
                    label="Quantum Efficiency:",
                    value=0.9,
                    min_value=0.01,
                    max_value=1.0,
                    key=f"qe_{i}",
                )
                settings[i]["pixelsize"] = col.number_input(
                    label="Pixelsize (nm):",
                    value=130,
                    min_value=10,
                    max_value=500,
                    key=f"pixelsize_{i}",
                )

                settings[i]["methods"] = col.selectbox(
                    "Method", options=["lq", "mle"], key=f"method_{i}"
                )

                settings[i]["undrift"] = col.number_input(
                    "Segmentation size for undrift",
                    value=1000,
                    min_value=100,
                    max_value=10000,
                    key=f"segmentation_{i}",
                )

                if col.checkbox("3D", key=f"3d_checkbox_{i}"):
                    col.info("3D will set fit method to lq-3d.")

                    settings[i]["methods"] = "lq-3d"

                    settings[i]["calib_file"] = col.text_input(
                        "Enter path to calibration file.", os.getcwd()
                    )

                    if not os.path.isfile(settings[i]["calib_file"]):
                        col.error("Not a valid file.")

                    settings[i]["magnification_factor"] = col.number_input(
                        label="Magnification factor:", value=0.79, key=f"magfac_{i}"
                    )
                else:
                    settings[i]["calib_file"] = None
                    settings[i]["magnification_factor"] = None

            if st.checkbox("Chain Custom command"):
                st.text("Allows to execute a custom command via shell.")
                st.text("You can pass the filename with $FILENAME.")
                command = st.text_input("Command", "")
            else:
                command = ""

            update_time = st.number_input(
                "Update time (scan every x-th minute):", DEFAULT_UPDATE_TIME
            )

            logfile_dir = os.path.dirname(localize._db_filename())
            now_str = datetime.now().strftime("%Y-%m-%d %H_%M_%S")
            logfile = os.path.join(logfile_dir, f"{now_str}_watcher.log")
            logfile = st.text_input("Logfile", logfile)

            if st.button("Submit"):

                settings_list = []

                for i in range(n_columns):
                    settings_selected = {}
                    settings_selected["box_side_length"] = settings[i]["box"]
                    settings_selected["gradient"] = settings[i]["min_net_gradient"]
                    settings_selected["gain"] = settings[i]["em_gain"]
                    settings_selected["baseline"] = settings[i]["baseline"]
                    settings_selected["sensitivity"] = settings[i]["sensitivity"]
                    settings_selected["qe"] = settings[i]["qe"]
                    settings_selected["pixelsize"] = settings[i]["pixelsize"]
                    settings_selected["fit_method"] = settings[i]["methods"]
                    settings_selected["drift"] = settings[i]["undrift"]
                    if settings[i]["calib_file"]:
                        settings_selected["zc"] = settings[i]["calib_file"]
                    if settings[i]["magnification_factor"]:
                        settings_selected["mf"] = settings[i]["magnification_factor"]

                    settings_selected["database"] = True
                    if n_columns == 1:
                        suffix = ""
                    else:
                        suffix = f"_pg_{i+1}"
                    settings_selected["suffix"] = suffix
                    settings_list.append(settings_selected)

                st.write(settings_list)

                existing = fetch_db()

                if len(existing) > 0:
                    existing = existing['filename'].tolist()
                else:
                    existing = []

                p = Process(
                    target=check_new_and_process,
                    args=(settings_list, folder, command, logfile, existing, update_time),
                )
                p.start()

                st.success("Started watcher.")
                display = st.empty()

                df = pd.DataFrame(
                    {
                        "process id": [p.pid],
                        "started": [datetime.now()],
                        "folder": [folder],
                        "logfile": [logfile],
                    }
                )

                engine = create_engine(
                    "sqlite:///" + localize._db_filename(), echo=False
                )
                df.to_sql("watcher", con=engine, if_exists="append", index=False)

                for reset in range(3):
                    display.success(f"Restarting in {3-reset}.")
                    time.sleep(1)
                display.success("Please refresh page.")
                st.stop()


"""
    picasso.render
    ~~~~~~~~~~~~~~

    Render single molecule localizations to a super-resolution image

    :author: Joerg Schnitzbauer, 2015
    :copyright: Copyright (c) 2015 Jungmann Lab, MPI of Biochemistry
"""
import time
import os
import sys

import numpy as _np
import numba as _numba
import scipy.signal as _signal
from scipy.spatial.transform import Rotation as _Rotation


_DRAW_MAX_SIGMA = 3


def render(
    locs,
    info=None,
    oversampling=1,
    viewport=None,
    blur_method=None,
    min_blur_width=0,
    ang=None,
):  
    """
    Renders locs.

    Parameters
    ----------
    locs : np.recarray
        Localizations to be rendered
    info : dict (default=None)
        Contains metadata for locs. Needed only if no viewport 
        specified
    oversampling : float (default=1)
        Number of super-resolution pixels per camera pixel
    viewport : list or tuple (default=None)
        Field of view to be rendered. If None, all locs are rendered
    blur_method : str (default=None)
        Defines localizations' blur. The string has to be one of 
        'gaussian', 'gaussian_iso', 'smooth', 'convolve'. If None, 
        no blurring is applied.
    min_blur_width : float (default=0)
        Minimum size of blur (pixels)
    ang : tuple (default=None)
        Rotation angles of locs around x, y and z axes. If None, 
        locs are not rotated.

    Raises
    ------
    Exception
        If blur_method not one of 'gaussian', 'gaussian_iso', 'smooth', 
        'convolve' or None

    Returns
    -------
    int
        Number of localizations rendered
    np.array
        Rendered image
    """

    if viewport is None:
        try:
            # all locs
            viewport = [(0, 0), (info[0]["Height"], info[0]["Width"])]
        except TypeError:
            raise ValueError("Need info if no viewport is provided.")
    (y_min, x_min), (y_max, x_max) = viewport
    if blur_method is None:
        # no blur
        return render_hist(
            locs, 
            oversampling, 
            y_min, x_min, y_max, x_max, 
            ang=ang, 
        )
    elif blur_method == "gaussian":
        # individual localization precision
        return render_gaussian(
            locs, 
            oversampling, 
            y_min, x_min, y_max, x_max, 
            min_blur_width, 
            ang=ang, 
        )
    elif blur_method == "gaussian_iso":
        # individual localization precision (same for x and y)
        return render_gaussian_iso(
            locs, 
            oversampling, 
            y_min, x_min, y_max, x_max, 
            min_blur_width, 
            ang=ang, 
        )
    elif blur_method == "smooth":
        # one pixel blur
        return render_smooth(
            locs, 
            oversampling, 
            y_min, x_min, y_max, x_max, 
            ang=ang, 
        )
    elif blur_method == "convolve":
        # global localization precision
        return render_convolve(
            locs, 
            oversampling, 
            y_min, x_min, y_max, x_max, 
            min_blur_width, 
            ang=ang, 
        )
    else:
        raise Exception("blur_method not understood.")


@_numba.njit
def _render_setup(
    locs, 
    oversampling, 
    y_min, x_min, y_max, x_max,
):
    """
    Finds coordinates to be rendered and sets up an empty image array.

    Parameters
    ----------
    locs : np.recarray
        Localizations
    oversampling : float
        Number of super-resolution pixels per camera pixel
    y_min : float
        Minimum y coordinate to be rendered (pixels)
    x_min : float
        Minimum x coordinate to be rendered (pixels)
    y_max : float
        Maximum y coordinate to be rendered (pixels)
    x_max : float
        Maximum x coordinate to be rendered (pixels)

    Returns
    -------
    np.array
        Empty image array
    int
        Number of pixels in y
    int 
        Number of pixels in x
    np.array 
        x coordinates to be rendered
    np.array 
        y coordinates to be rendered
    np.array
        Indeces of locs to be rendered
    """

    n_pixel_y = int(_np.ceil(oversampling * (y_max - y_min)))
    n_pixel_x = int(_np.ceil(oversampling * (x_max - x_min)))
    x = locs.x
    y = locs.y
    in_view = (x > x_min) & (y > y_min) & (x < x_max) & (y < y_max)
    x = x[in_view]
    y = y[in_view]
    x = oversampling * (x - x_min)
    y = oversampling * (y - y_min)
    image = _np.zeros((n_pixel_y, n_pixel_x), dtype=_np.float32)
    return image, n_pixel_y, n_pixel_x, x, y, in_view


@_numba.njit
def _render_setup3d(
    locs, 
    oversampling, 
    y_min, x_min, y_max, x_max, z_min, z_max, 
    pixelsize,
):
    """
    Finds coordinates to be rendered in 3D and sets up an empty image 
    array. Used by Picasso: Average3.

    Parameters
    ----------
    locs : np.recarray
        Localizations
    oversampling : float
        Number of super-resolution pixels per camera pixel
    y_min : float
        Minimum y coordinate to be rendered (pixels)
    x_min : float
        Minimum x coordinate to be rendered (pixels)
    y_max : float
        Maximum y coordinate to be rendered (pixels)
    x_max : float
        Maximum x coordinate to be rendered (pixels)
    z_min : float
        Minimum z coordinate to be rendered (nm)
    z_max : float
        Maximum z coordinate to be rendered (nm)
    pixelsize : float
        Camera pixel size, used for converting z coordinates

    Returns
    -------
    np.array
        Empty image array
    int
        Number of pixels in y
    int 
        Number of pixels in x
    int
        Number of pixels in z
    np.array 
        x coordinates to be rendered
    np.array 
        y coordinates to be rendered
    np.array
        z coordinates to be rendered
    np.array
        Indeces of locs to be rendered
    """

    n_pixel_y = int(_np.ceil(oversampling * (y_max - y_min)))
    n_pixel_x = int(_np.ceil(oversampling * (x_max - x_min)))
    n_pixel_z = int(_np.ceil(oversampling * (z_max - z_min)))
    x = locs.x
    y = locs.y
    z = locs.z / pixelsize
    in_view = (
        (x > x_min)
        & (y > y_min)
        & (z > z_min)
        & (x < x_max)
        & (y < y_max)
        & (z < z_max)
    )
    x = x[in_view]
    y = y[in_view]
    z = z[in_view]
    x = oversampling * (x - x_min)
    y = oversampling * (y - y_min)
    z = oversampling * (z - z_min)
    image = _np.zeros((n_pixel_y, n_pixel_x, n_pixel_z), dtype=_np.float32)
    return image, n_pixel_y, n_pixel_x, n_pixel_z, x, y, z, in_view

# @_numba.njit
# def _render_setupz(
#     locs, oversampling, x_min, z_min, x_max, z_max
# ):
#     n_pixel_x = int(_np.ceil(oversampling * (x_max - x_min)))
#     n_pixel_z = int(_np.ceil(oversampling * (z_max - z_min)))
#     x = locs.x
#     z = locs.z
#     in_view = (x > x_min) & (z > z_min) & (x < x_max) & (z < z_max)
#     x = x[in_view]
#     z = z[in_view]
#     x = oversampling * (x - x_min)
#     z = oversampling * (z - z_min)
#     image = _np.zeros((n_pixel_x, n_pixel_z), dtype=_np.float32)
#     return image, n_pixel_z, n_pixel_x, x, z, in_view


@_numba.njit
def _fill(image, x, y):
    """
    Fills image with x and y coordinates. 
    Image is not blurred.

    Parameters
    ----------
    image : np.array
        Empty image array
    x : np.array
        x coordinates to be rendered
    y : np.array
        y coordinates to be rendered
    """

    x = x.astype(_np.int32)
    y = y.astype(_np.int32)
    for i, j in zip(x, y):
        image[j, i] += 1


@_numba.njit
def _fill3d(image, x, y, z):
    """
    Fills image with x, y and z coordinates.
    Image is not blurred.
    Used by Picasso: Average3.

    Parameters
    ----------
    image : np.array
        Empty image array
    x : np.array
        x coordinates to be rendered
    y : np.array
        y coordinates to be rendered
    z : np.array
        z coordinates to be rendered
    """

    x = x.astype(_np.int32)
    y = y.astype(_np.int32)
    z = z.astype(_np.int32)
    z += _np.min(z) # because z takes also negative values
    for i, j, k in zip(x, y, z):
        image[j, i, k] += 1


@_numba.njit
def _fill_gaussian(image, x, y, sx, sy, n_pixel_x, n_pixel_y):
    """
    Fills image with blurred x and y coordinates.
    Localization precisions (sx and sy) are treated as standard
    deviations of the guassians to be rendered.

    Parameters
    ----------
    image : np.array
        Empty image array
    x : np.array
        x coordinates to be rendered
    y : np.array
        y coordinates to be rendered
    sx : np.array
        Localization precision in x for each loc
    sy : np.array
        Localization precision in y for each loc
    n_pixel_x : int
        Number of pixels in x
    n_pixel_y : int
        Number of pixels in y
    """

    # render each localization separately
    for x_, y_, sx_, sy_ in zip(x, y, sx, sy):

        # get min and max indeces to draw the given localization
        max_y = _DRAW_MAX_SIGMA * sy_
        i_min = _np.int32(y_ - max_y)
        if i_min < 0:
            i_min = 0
        i_max = _np.int32(y_ + max_y + 1)
        if i_max > n_pixel_y:
            i_max = n_pixel_y
        max_x = _DRAW_MAX_SIGMA * sx_
        j_min = _np.int32(x_ - max_x)
        if j_min < 0:
            j_min = 0
        j_max = _np.int32(x_ + max_x) + 1
        if j_max > n_pixel_x:
            j_max = n_pixel_x

        # draw a localization as a 2D guassian PDF
        for i in range(i_min, i_max):
            for j in range(j_min, j_max):
                image[i, j] += _np.exp(
                    -(
                        (j - x_ + 0.5) ** 2 / (2 * sx_**2)
                        + (i - y_ + 0.5) ** 2 / (2 * sy_**2)
                    )
                ) / (2 * _np.pi * sx_ * sy_)


@_numba.njit
def _fill_gaussian_rot(
    image, x, y, z, sx, sy, sz, n_pixel_x, n_pixel_y, ang
):
    """
    Fills image with rotated gaussian-blurred localizations.

    Localization precisions (sx, sy and sz) are treated as standard
    deviations of the guassians to be rendered.

    See https://cs229.stanford.edu/section/gaussians.pdf

    Parameters
    ----------
    image : np.array
        Empty image array
    x : np.array
        x coordinates to be rendered
    y : np.array
        y coordinates to be rendered
    z: np.array
        z coordinates to be rendered
    sx : np.array
        Localization precision in x for each loc
    sy : np.array
        Localization precision in y for each loc
    sz : np.array
        Localization precision in z for each loc
    n_pixel_x : int
        Number of pixels in x
    n_pixel_y : int
        Number of pixels in y
    ang : tuple
        Rotation angles of locs around x, y and z axes.
    """

    (angx, angy, angz) = ang # rotation angles

    rot_mat_x = _np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, _np.cos(angx), _np.sin(angx)],
            [0.0, -_np.sin(angx), _np.cos(angx)],
        ], dtype=_np.float32
    ) # rotation matrix around x axis
    rot_mat_y = _np.array(
        [
            [_np.cos(angy), 0.0, _np.sin(angy)],
            [0.0, 1.0, 0.0],
            [-_np.sin(angy), 0.0, _np.cos(angy)],
        ], dtype=_np.float32
    ) # rotation matrix around y axis
    rot_mat_z = _np.array(
        [
            [_np.cos(angz), -_np.sin(angz), 0.0],
            [_np.sin(angz), _np.cos(angz), 0.0],
            [0.0, 0.0, 1.0],
        ], dtype=_np.float32
    ) # rotation matrix around z axis
    # rot_matrix = _np.matmul(
    #     _np.matmul(
    #         rot_mat_x, rot_mat_y, dtype=_np.float32
    #     ), rot_mat_z, dtype=_np.float32
    # )
    rot_matrix = rot_mat_x @ rot_mat_y @ rot_mat_z # rotation matrix
    rot_matrixT = _np.transpose(rot_matrix) # ...and its transpose

    # draw each localization separately
    for x_, y_, z_, sx_, sy_, sz_ in zip(x, y, z, sx, sy, sz):

        # get min and max indeces to draw the given localization
        max_y = (_DRAW_MAX_SIGMA * 2.5) * sy_
        i_min = int(y_ - max_y)
        if i_min < 0:
            i_min = 0
        i_max = int(y_ + max_y + 1)
        if i_max > n_pixel_y:
            i_max = n_pixel_y
        max_x = (_DRAW_MAX_SIGMA * 2.5) * sx_
        j_min = int(x_ - max_x)
        if j_min < 0:
            j_min = 0
        j_max = int(x_ + max_x + 1)
        if j_max > n_pixel_x:
            j_max = n_pixel_x
        max_z = (_DRAW_MAX_SIGMA * 2.5) * sz_ 
        k_min = int(z_ - max_z)
        k_max = int(z_ + max_z + 1)

        # rotate localization precisions in 3D
        cov_matrix = _np.array(
            [
                [sx_**2, 0, 0], 
                [0, sy_**2, 0], 
                [0, 0, sz_**2],
            ], dtype=_np.float32
        ) # covariance matrix (CM)
        cov_rot = rot_matrix @ cov_matrix @ rot_matrixT # rotated CM
        cri = inverse_3x3(cov_rot) # inverse of rotated CM
        dcr = determinant_3x3(cov_rot) # determinant of rotated CM

        # draw a localization in 2D - sum z coordinates;
        # PDF of a rotated gaussian in 3D is calculated and 
        # image is summed over z axis
        for i in range(i_min, i_max):
            b = _np.float32(i + 0.5 - y_)
            for j in range(j_min, j_max):
                a = _np.float32(j + 0.5 - x_)
                for k in range(k_min, k_max):   
                    c = _np.float32(k + 0.5 - z_)
                    exponent = (
                        a*a * cri[0,0] + a*b * cri[0,1] + a*c * cri[0,2] 
                        + a*b * cri[1,0] + b*b * cri[1,1] + b*c * cri[1,2]
                        + a*c * cri[2,0] + b*c * cri[2,1] + c*c * cri[2,2]
                    ) # Mahalanobis distance
                    image[i,j] += (
                        _np.exp(-0.5 * exponent)
                        / (((2*_np.pi) ** 3 * dcr) ** 0.5)
                    )


@_numba.njit
def inverse_3x3(a):
    """
    Calculates inverse of a 3x3 matrix.

    This function is faster than np.linalg.inv.

    Parameters
    ----------
    a : np.array
        3x3 matrix

    Returns
    -------
    np.array
        Inverse of a
    """

    c = _np.zeros((3,3), dtype=_np.float32)
    det = determinant_3x3(a)

    c[0,0] = (a[1,1] * a[2,2] - a[1,2] * a[2,1]) / det
    c[0,1] = (a[0,2] * a[2,1] - a[0,1] * a[2,2]) / det
    c[0,2] = (a[0,1] * a[1,2] - a[0,2] * a[1,1]) / det

    c[1,0] = (a[1,2] * a[2,0] - a[1,0] * a[2,2]) / det
    c[1,1] = (a[0,0] * a[2,2] - a[0,2] * a[2,0]) / det
    c[1,2] = (a[0,2] * a[1,0] - a[0,0] * a[1,2]) / det

    c[2,0] = (a[1,0] * a[2,1] - a[1,1] * a[2,0]) / det
    c[2,1] = (a[0,1] * a[2,0] - a[0,0] * a[2,1]) / det
    c[2,2] = (a[0,0] * a[1,1] - a[0,1] * a[1,0]) / det

    return c    


@_numba.njit
def determinant_3x3(a):
    """
    Calculates determinant of a 3x3 matrix.

    This function is faster than np.linalg.det.

    Parameters
    ----------
    a : np.array
        3x3 matrix

    Returns
    -------
    float
        Determinant of a
    """

    return _np.float32(
        a[0,0] * (a[1,1] * a[2,2] - a[1,2] * a[2,1]) 
        - a[0,1] * (a[1,0] * a[2,2] - a[2,0] * a[1,2]) 
        + a[0,2] * (a[1,0] * a[2,1] - a[2,0] * a[1,1])
    )


def render_hist(
    locs, 
    oversampling, 
    y_min, x_min, y_max, x_max, 
    ang=None, 
):
    """
    Renders locs with no blur.

    Parameters
    ----------
    locs : np.recarray
        Localizations to be rendered
    oversampling : float (default=1)
        Number of super-resolution pixels per camera pixel
    y_min : float
        Minimum y coordinate to be rendered (pixels)
    x_min : float
        Minimum x coordinate to be rendered (pixels)
    y_max : float
        Maximum y coordinate to be rendered (pixels)
    x_max : float
        Maximum x coordinate to be rendered (pixels)
    ang : tuple (default=None)
        Rotation angles of locs around x, y and z axes. If None, 
        locs are not rotated.

    Returns
    -------
    int
        Number of localizations rendered
    np.array
        Rendered image
    """

    image, n_pixel_y, n_pixel_x, x, y, in_view = _render_setup(
        locs, 
        oversampling, 
        y_min, x_min, y_max, x_max,
    )
    if ang:
        x, y, _, _ = locs_rotation(
            locs, 
            oversampling,
            x_min, x_max, y_min, y_max, 
            ang, 
        )
    _fill(image, x, y)
    return len(x), image


# @_numba.jit(nopython=True, nogil=True)
# def render_histz(locs, oversampling, x_min, z_min, x_max, z_max):
#     image, n_pixel_z, n_pixel_x, x, z, in_view = _render_setupz(
#         locs, oversampling, x_min, z_min, x_max, z_max
#     )
#     _fill(image, z, x)
#     return len(x), image


@_numba.jit(nopython=True, nogil=True)
def render_hist3d(
    locs, 
    oversampling, 
    y_min, x_min, y_max, x_max, z_min, z_max, 
    pixelsize,
):
    """
    Renders locs in 3D with no blur.
    Used by Picasso: Average3.

    Parameters
    ----------
    locs : np.recarray
        Localizations to be rendered
    oversampling : float (default=1)
        Number of super-resolution pixels per camera pixel
    y_min : float
        Minimum y coordinate to be rendered (pixels)
    x_min : float
        Minimum x coordinate to be rendered (pixels)
    y_max : float
        Maximum y coordinate to be rendered (pixels)
    x_max : float
        Maximum x coordinate to be rendered (pixels)
    z_min : float
        Minimum z coordinate to be rendered (nm)
    z_max : float
        Maximum z coordinate to be rendered (nm)
    pixelsize : float
        Camera pixel size, used for converting z coordinates

    Returns
    -------
    int
        Number of localizations rendered
    np.array
        Rendered 3D image
    """

    z_min = z_min / pixelsize
    z_max = z_max / pixelsize

    image, n_pixel_y, n_pixel_x, n_pixel_z, x, y, z, in_view = _render_setup3d(
        locs, 
        oversampling, 
        y_min, x_min, y_max, x_max, z_min, z_max, 
        pixelsize,
    )
    _fill3d(image, x, y, z)
    return len(x), image


def render_gaussian(
    locs, 
    oversampling, 
    y_min, x_min, y_max, x_max, 
    min_blur_width, 
    ang=None, 
):
    """
    Renders locs with with individual localization precision which 
    differs in x and y.

    Parameters
    ----------
    locs : np.recarray
        Localizations to be rendered
    oversampling : float (default=1)
        Number of super-resolution pixels per camera pixel
    y_min : float
        Minimum y coordinate to be rendered (pixels)
    x_min : float
        Minimum x coordinate to be rendered (pixels)
    y_max : float
        Maximum y coordinate to be rendered (pixels)
    x_max : float
        Maximum x coordinate to be rendered (pixels)
    min_blur_width : float
        Minimum localization precision (pixels)
    ang : tuple (default=None)
        Rotation angles of locs around x, y and z axes. If None, 
        locs are not rotated.

    Returns
    -------
    int
        Number of localizations rendered
    np.array
        Rendered image
    """

    image, n_pixel_y, n_pixel_x, x, y, in_view = _render_setup(
        locs, 
        oversampling, 
        y_min, x_min, y_max, x_max,
    )

    if not ang: # not rotated
        blur_width = oversampling * _np.maximum(locs.lpx, min_blur_width)
        blur_height = oversampling * _np.maximum(locs.lpy, min_blur_width)
        sy = blur_height[in_view]
        sx = blur_width[in_view]
    
        _fill_gaussian(image, x, y, sx, sy, n_pixel_x, n_pixel_y)

    else: # rotated
        x, y, in_view, z = locs_rotation(
            locs, 
            oversampling, 
            x_min, x_max, y_min, y_max, 
            ang, 
        )
        blur_width = oversampling * _np.maximum(locs.lpx, min_blur_width)
        blur_height = oversampling * _np.maximum(locs.lpy, min_blur_width)
        # for now, let lpz be twice the mean of lpx and lpy (TODO):
        if hasattr(locs, "lpz"):
            lpz = locs.lpz #NOTE: lpz must be in the same units as lpx
        else:
            lpz = 2 * _np.mean(_np.stack((locs.lpx, locs.lpy)), axis=0)
        blur_depth = oversampling * _np.maximum(lpz, min_blur_width)

        sy = blur_height[in_view]
        sx = blur_width[in_view]
        sz = blur_depth[in_view]

        _fill_gaussian_rot(
            image, x, y, z, sx, sy, sz, n_pixel_x, n_pixel_y, ang
        )

    return len(x), image


def render_gaussian_iso(
    locs, 
    oversampling, 
    y_min, x_min, y_max, x_max, 
    min_blur_width, 
    ang=None, 
):
    """
    Renders locs with with individual localization precision which 
    is the same in x and y.

    Parameters
    ----------
    locs : np.recarray
        Localizations to be rendered
    oversampling : float (default=1)
        Number of super-resolution pixels per camera pixel
    y_min : float
        Minimum y coordinate to be rendered (pixels)
    x_min : float
        Minimum x coordinate to be rendered (pixels)
    y_max : float
        Maximum y coordinate to be rendered (pixels)
    x_max : float
        Maximum x coordinate to be rendered (pixels)
    min_blur_width : float
        Minimum localization precision (pixels)
    ang : tuple (default=None)
        Rotation angles of locs around x, y and z axes. If None, 
        locs are not rotated.

    Returns
    -------
    int
        Number of localizations rendered
    np.array
        Rendered image
    """

    image, n_pixel_y, n_pixel_x, x, y, in_view = _render_setup(
        locs, 
        oversampling, 
        y_min, x_min, y_max, x_max,
    )

    if not ang: # not rotated
        blur_width = oversampling * _np.maximum(locs.lpx, min_blur_width)
        blur_height = oversampling * _np.maximum(locs.lpy, min_blur_width)
        sy = (blur_height[in_view] + blur_width[in_view]) / 2
        sx = sy
    
        _fill_gaussian(image, x, y, sx, sy, n_pixel_x, n_pixel_y)

    else: # rotated
        x, y, in_view, z = locs_rotation(
            locs, 
            oversampling, 
            x_min, x_max, y_min, y_max, 
            ang, 
        )
        blur_width = oversampling * _np.maximum(locs.lpx, min_blur_width)
        blur_height = oversampling * _np.maximum(locs.lpy, min_blur_width)
        # for now, let lpz be twice the mean of lpx and lpy (TODO):
        if hasattr(locs, "lpz"):
            lpz = locs.lpz #NOTE: lpz must be in the same units as lpx
        else:
            lpz = 2 * _np.mean(_np.stack((locs.lpx, locs.lpy)), axis=0)
        blur_depth = oversampling * _np.maximum(lpz, min_blur_width)

        sy = (blur_height[in_view] + blur_width[in_view]) / 2
        sx = sy
        sz = blur_depth[in_view]

        _fill_gaussian_rot(
            image, x, y, z, sx, sy, sz, n_pixel_x, n_pixel_y, ang
        )

    return len(x), image


def render_convolve(
    locs, 
    oversampling, 
    y_min, x_min, y_max, x_max, 
    min_blur_width, 
    ang=None, 
):
    """
    Renders locs with with global localization precision, i.e. each
    localization is blurred by the median localization precision in x
    and y.

    Parameters
    ----------
    locs : np.recarray
        Localizations to be rendered
    oversampling : float (default=1)
        Number of super-resolution pixels per camera pixel
    y_min : float
        Minimum y coordinate to be rendered (pixels)
    x_min : float
        Minimum x coordinate to be rendered (pixels)
    y_max : float
        Maximum y coordinate to be rendered (pixels)
    x_max : float
        Maximum x coordinate to be rendered (pixels)
    min_blur_width : float
        Minimum localization precision (pixels)
    ang : tuple (default=None)
        Rotation angles of locs around x, y and z axes. If None, 
        locs are not rotated.

    Returns
    -------
    int
        Number of localizations rendered
    np.array
        Rendered image
    """

    image, n_pixel_y, n_pixel_x, x, y, in_view = _render_setup(
        locs, 
        oversampling, 
        y_min, x_min, y_max, x_max,
    )
    if ang: # rotate
        x, y, in_view, _ = locs_rotation(
            locs, 
            oversampling, 
            x_min, x_max, y_min, y_max, 
            ang, 
        ) 

    n = len(x)
    if n == 0:
        return 0, image
    else:
        _fill(image, x, y)
        blur_width = oversampling * max(
            _np.median(locs.lpx[in_view]), min_blur_width
        )
        blur_height = oversampling * max(
            _np.median(locs.lpy[in_view]), min_blur_width
        )
        return n, _fftconvolve(image, blur_width, blur_height)


def render_smooth(
    locs, 
    oversampling, 
    y_min, x_min, y_max, x_max, 
    ang=None, 
):
    """
    Renders locs with with blur of one display pixel (set by 
    oversampling)

    Parameters
    ----------
    locs : np.recarray
        Localizations to be rendered
    oversampling : float (default=1)
        Number of super-resolution pixels per camera pixel
    y_min : float
        Minimum y coordinate to be rendered (pixels)
    x_min : float
        Minimum x coordinate to be rendered (pixels)
    y_max : float
        Maximum y coordinate to be rendered (pixels)
    x_max : float
        Maximum x coordinate to be rendered (pixels)
    ang : tuple (default=None)
        Rotation angles of locs around x, y and z axes. If None, 
        locs are not rotated.

    Returns
    -------
    int
        Number of localizations rendered
    np.array
        Rendered image
    """

    image, n_pixel_y, n_pixel_x, x, y, in_view = _render_setup(
        locs, 
        oversampling, 
        y_min, x_min, y_max, x_max,
    )

    if ang:
        x, y, _, _ = locs_rotation(
            locs, 
            oversampling, 
            x_min, x_max, y_min, y_max, 
            ang, 
        )

    n = len(x)
    if n == 0:
        return 0, image
    else:
        _fill(image, x, y)
        return n, _fftconvolve(image, 1, 1)


def _fftconvolve(image, blur_width, blur_height): 
    """
    Blurs (convolves) 2D image using fast fourier transform.

    Parameters
    ----------
    image : np.array
        Image with renderd but not blurred locs
    blur_width : float
        Blur width
    blur_height
        Blur height

    Returns
    -------
    np.array
        Blurred image
    """

    kernel_width = 10 * int(_np.round(blur_width)) + 1
    kernel_height = 10 * int(_np.round(blur_height)) + 1
    kernel_y = _signal.gaussian(kernel_height, blur_height)
    kernel_x = _signal.gaussian(kernel_width, blur_width)
    kernel = _np.outer(kernel_y, kernel_x)
    kernel /= kernel.sum()
    return _signal.fftconvolve(image, kernel, mode="same")


def rotation_matrix(angx, angy, angz):
    """
    Finds rotation matrix given rotation angles around axes.

    Parameters
    ----------
    angx : float
        Rotation angle around x axis
    angy : float
        Rotation angle around y axis
    angz : float
        Rotation angle around z axis

    Returns
    -------
    scipy.spatial.transform.Rotation
        Scipy class that can be applied to rotate an Nx3 np.array
    """

    rot_mat_x = _np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, _np.cos(angx), _np.sin(angx)],
            [0.0,-_np.sin(angx), _np.cos(angx)],
        ]
    ) # rotation matrix around x axis
    rot_mat_y = _np.array(
        [
            [_np.cos(angy), 0.0, _np.sin(angy)],
            [0.0, 1.0, 0.0],
            [-_np.sin(angy), 0.0, _np.cos(angy)],
        ]
    ) # rotation matrix around y axis
    rot_mat_z = _np.array(
        [
            [_np.cos(angz), -_np.sin(angz), 0.0],
            [_np.sin(angz), _np.cos(angz), 0.0],
            [0.0, 0.0, 1.0],
        ]
    ) # rotation matrix around z axis
    rot_mat = rot_mat_x @ rot_mat_y @ rot_mat_z
    return _Rotation.from_matrix(rot_mat)


def locs_rotation(
    locs, 
    oversampling,
    x_min, x_max, y_min, y_max, 
    ang, 
):
    """
    Rotates localizations within a FOV.

    Parameters
    ----------
    locs : np.recarray
        Localizations to be rotated
    oversampling : float
        Number of super-resolution pixels per camera pixel
    y_min : float
        Minimum y coordinate to be rendered (pixels)
    x_min : float
        Minimum x coordinate to be rendered (pixels)
    y_max : float
        Maximum y coordinate to be rendered (pixels)
    x_max : float
        Maximum x coordinate to be rendered (pixels)
    ang : tuple
        Rotation angles of locs around x, y and z axes.

    Returns
    -------
    np.array
        New (rotated) x coordinates
    np.array
        New y coordinates
    np.array
        Indeces of locs that are rendered
    np.array
        New z coordinates
    """

    # z is translated to pixels
    locs_coord = _np.stack((locs.x ,locs.y, locs.z)).T

    # x and y are in range (x_min/y_min, x_max/y_max) so they need to be
    # shifted (scipy rotation is around origin)
    locs_coord[:, 0] -= x_min + (x_max - x_min) / 2
    locs_coord[:, 1] -= y_min + (y_max - y_min) / 2

    # rotate locs
    R = rotation_matrix(ang[0], ang[1], ang[2])
    locs_coord = R.apply(locs_coord)

    # unshift locs
    locs_coord[:,0] += x_min + (x_max - x_min) / 2
    locs_coord[:,1] += y_min + (y_max - y_min) / 2

    # output
    x = locs_coord[:,0]
    y = locs_coord[:,1]
    z = locs_coord[:,2]
    in_view = (x > x_min) & (y > y_min) & (x < x_max) & (y < y_max)
    x = x[in_view]        
    y = y[in_view]
    z = z[in_view] 
    x = oversampling * (x - x_min)
    y = oversampling * (y - y_min)
    z *= oversampling
    return x, y, in_view, z
